
' AdvancedPassthrough
'
' Using Jet 2.0, you can use quickly and easily retrieve
' multiple result sets and warnings and/or print statements
' from stored procedures.
' 
' See the MS Access 2.0 help topic "Creating a Passthrough Query"
' for more information on the details of what is presented here.
'
Sub advancedpassthrough ()

' Scenario: You have a stored procedure that returns
' multiple results sets that you need to use.

' Solution: Using Access 2.0, first create a query that
' executes the stored procedure.  Open the "pubs.mdb"
' database in Access and examine the query called:
'   "Run: Multiple Result Set Sproc"
' To see and/or create the corresponding sproc on SQL
' server, look at its companion passthrough query:
'   "Create: Multiple Result Set Sproc"
'
' You cannot create stored Jet passthrough queries from VB3
' because DAO lacks the necessary properties (it is
' possible in Access 2' DAO).

' If you simply execute the stored procedure, you will get
' a snapshot with the first result set returned.

Dim d As Database
Set d = OpenDatabase("pubs.mdb")

Dim s As Snapshot

Set s = d.CreateSnapshot("Run: Multiple Result Set Sproc")

Debug.Print s(0).Name, s(1).Name, s(2).Name
Debug.Print s(0), s(1), s(2)

' Depending on your scenario, that may be all you need.
' However, you may wish to get the other result sets.
' You can do this by running a maketable query version
' of the stored procedure as follows:

ResultTable = "Results"
d.Execute "SELECT * INTO " & ResultTable & " FROM [Run: Multiple Result Set Sproc];"

' This will create a set of (local) tables,
'       Results, Results1 ...
' with each table containing successive returned result
' sets.  We can see this by listing them:

Debug.Print d.TableDefs(ResultTable).DateCreated
Debug.Print d.TableDefs(ResultTable & "1").DateCreated

' These can then be opened like any normal table or bound
' to a data control or whatever.
' When done with them, you can delete them in preparation
' for running the stored proc again.

d.TableDefs.Delete ResultTable
d.TableDefs.Delete ResultTable & "1"

' Another thing that stored procedures sometimes do is
' use the Transact SQL "Print" command to return misc
' data.  You can use those from Jet 2.0 also.  The msgs
' will get written to a new table called "<username> - N".
' Thus:

Set s = d.CreateSnapshot("Run: Sproc w/ Print Statements")

' (Note: created with "Create: Sproc w/ Print Statements")

d.TableDefs.Refresh
Debug.Print d.TableDefs("Admin - 00").DateCreated

' Note that the table is only created on demand.  If the
' print statement didn't occur until after all the results
' had been returned, then we would have to close the
' snapshot before they appeared.

s.Close
d.Close

End Sub

' Minimize Connections
'
' When working with ODBC data,  it can help to be aware of your
' connection usage.  Although Jet quietly manages these for
' you, you can tweak its handling with some of the tips below.
'
'   + Preconnect at app startup to save time later
'   + limit returned dynasets to 100 records
'   + force connection release by completing dynaset filling
'   + modify the ConnectionTimeout in [ODBC] section
'   + Remember, connections can't be closed if:
'      - a transaction is pending
'      - not all query results have been fetched
'
Sub MinimizeConnections ()

' When you first reference an attached table, Jet will make a
' connection to the ODBC database.  Since creating the
' connection can be an expensive operation which causes a
' pause at a bad time, you may wish to force it to be opened
' at some earlier time.  The following code will force this:

Dim d As Database, ConnStr As String

ConnStr = "ODBC;"       ' prompt the user for info
Set d = OpenDatabase("", False, False, ConnStr)
ConnStr = d.Connect     ' save the filled in string for later
d.Close

' Note that your application should not use the database that
' you've opened above -- it should use attached tables.  The
' connection created above will last until it is timed out
' (default is 10 mins).

End Sub

' Use parameterized queries instead of 'dynamic SQL'.  Why?
'   + they execute faster
'   + they're easier to use than building SQL strings
'
Sub UseParameterizedQueries ()

' Say you wished to repeatedly find a set of records based
' on certain criteria.  For example, say you wanted all
' Authors whose name started with a certain letter.  One way
' to do this is to have code like:

Dim d As Database, ds As dynaset
Set d = OpenDatabase("biblio.mdb")
firstletter = "G"

sqlstr = "Select * from Authors where Author like """ & firstletter & "*"""

Set ds = d.CreateDynaset(sqlstr)
Debug.Print ds!Author
ds.Close

'A preferred way is to have a parameterized query already created
'and just set the parameter appropriately.  For example:

'I'll create the query here so that you can use standard biblio.
'Normally it would already exist in the database

Dim qd As querydef
Set qd = d.CreateQueryDef("AuthorLike", "PARAMETERS FirstLetter Text; SELECT DISTINCTROW Authors.Au_ID, Authors.Author FROM Authors WHERE ((Authors.Author Like [FirstLetter]))")
qd.Close

' This is where the code starts that you would normally write

Set qd = d.OpenQueryDef("AuthorLike")
qd!firstletter = firstletter & "*"
Set ds = qd.CreateDynaset()

Debug.Print ds!Author
ds.Close

' As you can see, the code is shorter, easier to understand,
' easier to maintain -- and, actually runs faster as well!

qd.Close                            ' tidy up after demo
d.DeleteQueryDef ("AuthorLike")
End Sub

' Its best to use SQL statements whereever possible rather
' than explicit coding.  Why?
'  + the query engine has tighter binding to the actual isam
'    than VB code
'  + the query engine has access to certain optimized operations
'    such as bulk copy operations
'  + SQL statements are quicker and easier to debug than the
'    equivalent code
'
Sub usesqlstatements ()

' A simple example is the following:

' bad code
Dim d As Database
Dim t As table

Set d = OpenDatabase("biblio.mdb")
Set t = d.OpenTable("Titles")

BeginTrans      ' faster with a transaction around it
While Not t.EOF
    t.Edit
    t![Year Published] = t![Year Published] - 100
    t.Update
    t.MoveNext
Wend
Rollback        ' added to keep the sample data intact
  
' Good code
BeginTrans
    d.Execute "Update Titles Set [Year Published] = [Year Published] - 100"
Rollback    ' only rolling back so the demo works repeatedly

' There are some caveats to note.
'
'   If an error occurs on an individual record, the rest of
'   the updates will still occur and you won't know that
'   an error occured.  In Access 2, this problem is avoided
'   by using the DB_FAILONERROR flag as an Execute option.  To
'   detect it in VB3, you need to get creative with things like:

Dim dsTemp As dynaset

BeginTrans      ' This is so I can rollback to maintain the data

    BeginTrans
    d.Execute "Update Titles Set [Year Published] = [Year Published] - 100"
    Set dsTemp = d.CreateDynaset("Select count(*) from Titles where [Year Published] > 1950")
    If dsTemp(0) <> 0 Then
        Rollback
    Else
        CommitTrans
    End If

Rollback        ' Again, this line not normally used!

' What I did above was to measure the success of the Execute.
' If the count(*) had come back > 0, then I would have known
' that one of the records hadn't been updated, due say, to it
' being locked by someone else.  Obviously you'll have to get
' creative for your own application.  Since Count(*) tends to
' be very fast, especially in Jet 2.0, this probably still
' works out faster than doing explicit coding.

' Another caveat is that you don't know how many records were
' modified, which is something that is often useful to give
' to users.  Again you can solve this with some extra SQL
' queries, and again, Access 2 solves this with the new
' "RowsAffected" property that you can read after the .Execute
' method completes.

End Sub

' It is best to use transactions where possible.  Why?
'
'   + To decrease programming complexity
'   + Around bulk updates to increase speed
'   + Around query result sets to increase concurrency
'
Sub UseTransactions ()

' + To decrease programming complexity
' ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
' In databases without transactions, its common to simulate
' them by making extensive use of temporary tables and variables.
' This just isn't as necessary in Jet.  By combining several
' operations within a transaction, you can abort the whole
' group just by issuing a rollback.

' For example, the following code adds a book to the database
' that has a new author.  Rather than get the author's
' information one at a time from the user and then store it
' in a temp table, similarly get the title and then confirm
' with the user that everything is ok and then make the
' actual changes, these can all be done immediately and
' then aborted if need be.

Dim d As Database, tAuthors As table, tTitles As table
Dim au_id As Long

Set d = OpenDatabase("biblio.mdb")
Set tAuthors = d.OpenTable("Authors")
Set tTitles = d.OpenTable("Titles")

BeginTrans      ' indented code to show transaction

    tAuthors.AddNew
    tAuthors!Author = "Austen, Jane"
    tAuthors.Update
    tAuthors.Bookmark = tAuthors.LastModified
    au_id = tAuthors!au_id
    
    tTitles.AddNew
    tTitles!Title = "Pride and Prejudice"
    tTitles![Year Published] = 1845
    tTitles!au_id = au_id1
    tTitles!ISBN = "123-4567-12345"
    tTitles.Update

' at this stage, before completing the transaction, you can
' look at the database and confirm that the data appears there
' however, by executing the following statement, it will be
' automatically removed from both tables.  This can be shown
' by printing the last record of the table before and after the
' rollback as follows:

tAuthors.MoveLast
Debug.Print tAuthors!Author     ' prints "Austen, Jane"

Rollback

tAuthors.MoveLast
Debug.Print tAuthors!Author     ' prints "Wiederhold, Gio."


' You can also nest transactions so that portions of an
' operation can be aborted without affecting other operations.
' For example you could confirm and commit the author and
' title information separately in nested transactions and still
' rollback the containing transaction - as follows:

BeginTrans      ' outer transaction

    BeginTrans
        tAuthors.AddNew
        tAuthors!Author = "Austen, Jane"
        tAuthors.Update
        tAuthors.Bookmark = tAuthors.LastModified
        au_id = tAuthors!au_id
    CommitTrans
    
    BeginTrans
        tTitles.AddNew
        tTitles!Title = "Pride and Prejudice"
        tTitles![Year Published] = 1845
        tTitles!au_id = au_id1
        tTitles!ISBN = "123-4567-12345"
        tTitles.Update
    CommitTrans

Rollback

' Although this may not seem especially useful in this trivial
' example, you can imagine how nested independent subroutines
' would find this technique extremely powerful.

' + Around bulk updates to increase speed
' ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
' If you're doing a large number of updates, you can group
' the updates within a transaction for speed.  Without
' transactions, every update is written to disk when its
' completed.  Updates inside a transaction are only written
' to disk when the transaction is completed.
'
' Note carefully the tradeoff however.  After the number of
' uncommitted transactions exceeds available memory, they will
' start being spooled to temporary storage on disk.  When you
' finally commit, they will first have to be read and deleted
' from temporary storage and then written to the actual tables.
'
' Of course the best way to do bulk updates is to use SQL
' statements like UPDATE and INSERT INTO (which already have
' internal transactions, so there's no benefit in placing them
' inside a transaction.

' + Around query result sets to increase concurrency
' ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
' Whenever you run a query that reads an index to find results,
' so-called "read locks" will be placed on that index so that
' other users don't update the locks beneath the query scan.
' For performance reasons, those locks time out during idle
' time (or when FreeLocks is called).  However, you can force
' faster timeout of the readlocks by writing code as follows:

Dim dsResults As dynaset
BeginTrans
    Set dsResults = d.CreateDynaset("By State")
CommitTrans

While Not dsResults.EOF
   ' do something
   dsResults.MoveNext
Wend

' When you are using the database single user, this will actually
' slow down your program.  Similarly if its unlikely that anyone
' else be writing to the same table as you're querying, you should
' also not do this.  However, if you know that other people will
' be updating the table that you are running a query against,
' this will unblock them faster.

End Sub

